#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import pathlib
from collections import Counter, defaultdict
from typing import Dict

from delivery.destinations.interfaces import (
    DeliveryContextIF,
    Destination,
    DestinationTempFile,
)
from delivery.finder import RestoreProductFinder
from delivery.products import Calibration, ProductMetadata, RestoreProduct

TEST_RESTORE_METADATA = ProductMetadata("EVLA", "17B-403", "restored_cms", None, None, "60468.6349189817")


class FakeDestination(Destination):
    def __init__(self):
        super().__init__(None)
        self.files_added: dict[str, pathlib.Path] = {}

    def add_file(self, file: pathlib.Path, relative_path: str):
        self.files_added[relative_path] = file

    def create_file(self, relative_path: str) -> DestinationTempFile:
        raise NotImplementedError

    def results(self) -> Dict:
        return {}


def test_web_calibration(resource_path_root):
    calibration = Calibration(resource_path_root / "calibrations" / "foo")

    # first, let's make sure we parsed the metadata properly
    project_code, pipeline_id = calibration.parse_metadata()
    assert project_code == "21A-409"
    assert pipeline_id == "59311.953874212966"

    # OK now let's make sure that we're delivering properly
    destination = FakeDestination()
    calibration.deliver_to(destination)

    expected_files = set(
        [
            "21A-409/calibration_pipeline.59311.953874212966/pipeline-20210409T164741/.keep",
            "21A-409/calibration_pipeline.59311.953874212966/unknown.hifv_cal.pipeline_manifest.xml",
            "21A-409/calibration_pipeline.59311.953874212966/casa_commands.log",
            "21A-409/calibration_pipeline.59311.953874212966/flux.csv",
            "21A-409/calibration_pipeline.59311.953874212966/unknown.hifv_cal.pprequest.xml",
            "21A-409/calibration_pipeline.59311.953874212966/21A-409.sb39530507.eb39564922.59311.953874212966.ms.calapply.txt",
            "21A-409/calibration_pipeline.59311.953874212966/casa_pipescript.py",
            "21A-409/calibration_pipeline.59311.953874212966/casa_piperestorescript.py",
            "21A-409/calibration_pipeline.59311.953874212966/PPR.xml",
            "21A-409/calibration_pipeline.59311.953874212966/pipeline_aquareport.xml",
            "21A-409/calibration_pipeline.59311.953874212966/weblog.tgz",
        ]
    )
    actual_files = destination.files_added.keys()

    # did we get all the expected files?
    for file in expected_files:
        assert file in actual_files

    # did we not get any unexpected files?
    for file in actual_files:
        assert file in expected_files


def test_restore(resource_path_root):
    restore = RestoreProduct(resource_path_root / "restores" / "tmpfpjeptku", TEST_RESTORE_METADATA)

    # OK now let's make sure that we're delivering properly
    destination = FakeDestination()
    restore.deliver_to(destination)

    deliver_rel_path_root = f"{TEST_RESTORE_METADATA.project_code}/{TEST_RESTORE_METADATA.pipeline_spec}"
    # Just check the file counts in each dir since they contain a lot of files
    expected_dirs_to_file_counts = Counter(
        {
            f"{deliver_rel_path_root}/17B-403.sb34574962.eb34577590_000.58035.544146898144.ms": 119,
            f"{deliver_rel_path_root}/weblog_calibration": 1495,
            f"{deliver_rel_path_root}/weblog_restore": 90,
        }
    )
    expected_files = {
        f"{deliver_rel_path_root}/PPR_restore.xml",
        f"{deliver_rel_path_root}/PPR_restore.xml",
        f"{deliver_rel_path_root}/17B-403.sb34574962.eb34577590_000.58035.544146898144.ms.calapply.txt",
        f"{deliver_rel_path_root}/casa_piperestorescript.py",
        f"{deliver_rel_path_root}/casa_pipescript.py",
        f"{deliver_rel_path_root}/casa_commands.log",
        f"{deliver_rel_path_root}/PPR_calibration.xml",
        f"{deliver_rel_path_root}/flux.csv",
        f"{deliver_rel_path_root}/unknown.session_1.caltables.tgz",
    }
    actual_files = destination.files_added

    # did we get all the expected files?
    for file in expected_files:
        assert file in actual_files
    actual_dirs_to_file_counts = Counter()
    actual_dirs_to_files = defaultdict(list)
    for filename, filepath in actual_files.items():
        for dir, _ in expected_dirs_to_file_counts.items():
            if filename.startswith(f"{dir}/") and filepath.is_file():
                actual_dirs_to_file_counts[dir] += 1
                actual_dirs_to_files[dir].append(filename)
    pathlib.Path("actual_dirs").write_text(str(actual_dirs_to_files.items()))
    assert actual_dirs_to_file_counts == expected_dirs_to_file_counts

    # did we not get any unexpected files?
    for file in actual_files:
        assert file in expected_files or any(
            file.startswith(expected_dir) for expected_dir in expected_dirs_to_file_counts.keys()
        )


def test_restore_product_finder_parse_metadata(resource_path_root):
    expected_metadata = TEST_RESTORE_METADATA
    finder = RestoreProductFinder(resource_path_root / "restores" / "tmpfpjeptku")
    actual_metadata = finder.parse_metadata()
    assert actual_metadata == expected_metadata
